import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.Stack;
import java.util.List;
import java.util.ArrayList;

/**
 * A pile of cards - these piles must be cleared to win the game,
 * as well as the cards in the deck. Piles can be built up from
 * king to ace in alternating colours. Only a king can be placed
 * on an empty pile.
 * 
 * @author Michael Berry
 * @version 23/04/09
 */
public class Pile extends Actor implements Accepter
{

    /** The cards in the pile. */
    private Stack<Card> cards;
    /** Whether an empty pile has to start with a king */
    public static final boolean FORCE_KING = false;
    
    /**
     * Create a new pile with a list of cards.
     * @param cards the list of cards to create the pile from.
     */
    public Pile(Stack<Card> cards) {
        this.cards = cards;
        for(Card card : cards) {
            card.setAccepter(this);
        }
    }
    
    /**
     * Add all the cards that are on the pile, and
     * make sure all of them are face down except
     * the last.
     */
    public void addedToWorld(World world) {
        setImage("images/cards/empty.png");
        GreenfootImage image = new GreenfootImage(getImage().getWidth(), getImage().getHeight()+200);
        image.drawImage(new GreenfootImage("images/cards/empty.png"), 0, 0);
        setImage(image);
        for(int i=0 ; i<cards.size() ; i++) {
            Card card = cards.get(i);
            if(i==cards.size()-1) {
                card.setFlipped(false);
                card.setDraggable(true);
            }
            else {
                card.setFlipped(true);
                card.setDraggable(false);
            }
            getWorld().addObject(card, getX(), getY()-100+i*5);
        }
    }
    
    /**
     * Add a card onto the pile if it's legal.
     * @return true if the card has been added, false if it's been rejected.
     */
    public boolean addCard(Card card) {
        if(cards.size()==0 && (card.getValue()==Card.Value.KING || !FORCE_KING)) {
            add(card);
            return true;
        }
        if(cards.size()==0) {
            return false;
        }
        Card topCard = cards.peek();
        if((topCard.getSuit()==Card.Suit.HEARTS || topCard.getSuit()==Card.Suit.DIAMONDS) && (card.getSuit()==Card.Suit.HEARTS || card.getSuit()==Card.Suit.DIAMONDS)) return false;
        if((topCard.getSuit()==Card.Suit.CLUBS || topCard.getSuit()==Card.Suit.SPADES) && (card.getSuit()==Card.Suit.CLUBS || card.getSuit()==Card.Suit.SPADES)) return false;
        if(topCard.getValue().ordinal()-1==card.getValue().ordinal()) {
            add(card);
            return true;
        }
        return false;
    }
    
    /**
     * Get a list of cards on top of a certain card
     */
    public List<Card> getCardsOn(Card card) {
        if(! cards.contains(card)) return null;
        List<Card> cardsOn = new ArrayList<Card>();
        int index = cards.indexOf(card)+1;
        int max = cards.size();
        for(int i=index ; i<max ; i++) {
            cardsOn.add(cards.get(i));
        }
        return cardsOn;
    }
    
    /**
     * Add a card to the pile of cards.
     */
    private void add(Card card) {
        if(card.getAccepter()!=null) {
            ((Pile)(card.getAccepter())).remove();
        }
        card.setAccepter(this);
        card.setLocation(getX(), getY()-100+getOffset());
        cards.push(card);
        card.setDraggable(true);
    }
    
    /**
     * Get the top card on the pile
     */
    public Card peek() {
        return cards.peek();
    }
    
    /**
     * Remove the top card from the pile
     */
    public void remove() {
        cards.pop();
        if(cards.size()>0) {
            peek().setFlipped(false);
            peek().setDraggable(true);
        }
    }
    
    /**
     * Get the vertical offset needed when adding a new card.
     */
    private int getOffset() {
        int offset = 0;
        for(Card card : cards) {
            if(card.isFlipped()) {
                offset += 5;
            }
            else {
                offset += 15;
            }
        }
        return offset;
    }

}
